<?php

namespace Drupal\monarch_csv_import;

use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Database\Connection;
use Drupal\Core\File\FileSystemInterface;

/**
 * Handles importing CSVs into database tables.
 */
class CsvImport {

  const DEFAULT_MAX_SIZE = 16;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * Constructs a CsvImport object.
   *
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   */
  public function __construct(FileSystemInterface $file_system, Connection $database) {
    $this->fileSystem = $file_system;
    $this->database = $database;
  }

  /**
   * Import CSV file into a table.
   *
   * @param string $csv_path
   *   The file name. Can be absolute path or public/private uri.
   * @param string|null $table_name
   *   The table name to import into. If no name is specified, then a name will
   *   be generated based on $csv_path. The table name is auto-prefixed with
   *   'csv_import_' to prevent clashes with Drupal's own tables. If the table
   *   exists it will be dropped and recreated.
   * @param array|null $primary_key_columns
   *   The columns that represent the primary key. If no columns are specified
   *   Then the first column is used.
   * @param array $index_columns
   *   The column names to flag as indexes.
   * @param array $banned_columns
   *   Fields to exclude from the import.
   */
  public function toTable(string $csv_path, ?string $table_name, ?array $primary_key_columns, array $index_columns = [], array $banned_columns = []) {
    $csv = $this->toIterable($csv_path);

    $header = $csv->columns();

    if (empty($table_name)) {
      $pathinfo = pathinfo($csv_path);
      $table_name = 'csv_import_' . preg_replace('/[\W_]+/', '_', $pathinfo['dirname'] . '_' . $pathinfo['filename']);
    }
    else {
      if (!preg_match('/^csv_import_/', $table_name)) {
        $table_name = 'csv_import_' . $table_name;
      }
    }

    if (empty($primary_key_columns)) {
      $primary_key_columns = [reset($header)];
    }

    if (empty($header)) {
      throw new \Exception('The first line of ' . $csv_path . ' is blank, but this should be a list of the column names.');
    }

    $lines = [];

    $max_size = [];

    foreach ($csv as $line) {
      foreach ($line as $key => $value) {
        if (is_array($value) || is_object($value)) {
          $value = $line[$key] = json_encode($value);
        }
        elseif (!is_string($value)) {
          $value = $line[$key] = (string) $value;
        }

        $max_size[$key] = max($max_size[$key] ?? static::DEFAULT_MAX_SIZE, strlen($value));
      }

      foreach ($banned_columns as $banned_column) {
        unset($line[$banned_column]);
      }

      $lines[] = $line;
    }

    $header = array_unique(array_filter($header, function ($value) use ($banned_columns) {
      return !in_array($value, $banned_columns);
    }));

    $table_definition = [
      'primary key' => $primary_key_columns,
      'indexes' => [],
      'fields' => [],
    ];

    foreach ($index_columns as $index => $field) {
      if (is_string($field)) {
        $table_definition['indexes'][$index][] = $field;
      }
      elseif (is_array($field)) {
        $table_definition['indexes'][$index] = $field;
      }
    }

    foreach ($header as $index => $field) {
      $table_definition['fields'][$field] = [
        'type' => 'varchar',
        // Align max size to the next power of 2.
        'length' => pow(2, ceil(log(($max_size[$field] ?? static::DEFAULT_MAX_SIZE) + 1, 2))),
        'not null' => in_array($field, $primary_key_columns),
        'serialize' => FALSE,
      ];
    }

    if ($this->database->schema()->tableExists($table_name)) {
      $this->database->schema()->dropTable($table_name);
    }

    $this->database->schema()->createTable($table_name, $table_definition);

    $insert_query = $this->database->insert($table_name)->fields($header);

    foreach ($lines as $line) {
      $insert_query->values($line);
    }

    $insert_query->execute();
  }

  /**
   * Create an iterable from a CSV file.
   *
   * @param string $csv_path
   *   The file name. Can be absolute path or public/private uri.
   * @param string $separator
   *   Set the field delimiter (one character only).
   * @param string $enclosure
   *   Set the field enclosure character (one character only).
   * @param string $escape
   *   Set the escape character (one character only). Defaults as a backslash.
   *
   * @return CsvIterator
   *   The iterator for the specified csv.
   */
  public function toIterable(string $csv_path, string $separator = ',', string $enclosure = '"', string $escape = '\\') : CsvIterator {
    $file_path = $this->fileSystem->realpath($csv_path);

    if (!$file_path) {
      throw new \Exception('File ' . $csv_path . ' does not exist.');
    }

    return new CsvIterator($file_path, $separator, $enclosure, $escape);
  }

  /**
   * Create an iterable from a CSV file.
   *
   * @param string $csv_path
   *   The csv file name. Can be absolute path or public/private uri.
   * @param string $yaml_path
   *   The yaml file name. Can be absolute path or public/private uri.
   * @param string $separator
   *   Set the field delimiter (one character only).
   * @param string $enclosure
   *   Set the field enclosure character (one character only).
   * @param string $escape
   *   Set the escape character (one character only). Defaults as a backslash.
   */
  public function toYaml(string $csv_path, string $yaml_path, string $separator = ',', string $enclosure = '"', string $escape = '\\') {
    $file_path = $this->fileSystem->realpath($csv_path);
    $out_path = $this->fileSystem->realpath($yaml_path);

    if (!$file_path) {
      throw new \Exception('File ' . $csv_path . ' does not exist.');
    }

    $data = iterator_to_array(new CsvIterator($file_path, $separator, $enclosure, $escape));
    $data = Yaml::encode($data);

    if (empty($data)) {
      throw new \Exception('No data to write.');
    }

    if (file_put_contents($out_path, $data) === FALSE) {
      throw new \Exception('Unable to write: ' . $yaml_path);
    }
  }

  /**
   * Create an iterable from a CSV file.
   *
   * @param string $csv_path
   *   The csv file name. Can be absolute path or public/private uri.
   * @param string $json_path
   *   The json file name. Can be absolute path or public/private uri.
   * @param string $separator
   *   Set the field delimiter (one character only).
   * @param string $enclosure
   *   Set the field enclosure character (one character only).
   * @param string $escape
   *   Set the escape character (one character only). Defaults as a backslash.
   */
  public function toJson(string $csv_path, string $json_path, string $separator = ',', string $enclosure = '"', string $escape = '\\') {
    $file_path = $this->fileSystem->realpath($csv_path);
    $out_path = $this->fileSystem->realpath($json_path);

    if (!$file_path) {
      throw new \Exception('File ' . $csv_path . ' does not exist.');
    }

    $data = iterator_to_array(new CsvIterator($file_path, $separator, $enclosure, $escape));
    $data = json_encode($data, JSON_PRETTY_PRINT);

    if (empty($data)) {
      throw new \Exception('No data to write.');
    }

    if (file_put_contents($out_path, $data) === FALSE) {
      throw new \Exception('Unable to write: ' . $json_path);
    }
  }

}
